package HSPC::ErrorStack;
use SWsoft::protect; ## automatic


=head1 NAME

B<HSPC::ErrorStack> - object-oriented error handling routines.

=head1 SYNOPSYS

 use HSPC::ErrorStack;

 $ENV{error_obj} = new HSPC::ErrorStack;
 $ENV{error_obj}->set(mod => MOD_CP, smod => APACHE, err => 1);
 $error = $ENV{error_obj}->get_last_error();
 print $ENV{error_obj}->get_stack_as_text();

=head1 DESCRIPTION

Package B<HSPC::ErrorStack> used for manipulations with error stack from all HSPC
applications written in perl: set error codes, read error codes,
get text description for error, etc.

=head1 USAGE

This chapter describes usage of B<HSPC::ErrorStack> in HSPC web applications.

=head2 INITIALIZATION AND ACCESS TO ERROR STACK

At startup (in I<web_init()>) should be created object B<HSPC::ErrorStack> and 
set I<$ENV{error_obj}> as reference to created object:

 $ENV{error_obj} = new HSPC::ErrorStack;

Any application that needed to set error should use method I<set_last_error()> 
or simply I<set()> (it's synonym), for example:
 
 $ENV{error_obj}->set(mod => MOD_CP, smod => APACHE, err => 12)
 
Note: I<module_id> should be constant from B<HSPC::ErrorStack>. 
Constants dynamically loaded from database at module startup.

For reading error code use method I<get_last_error()> or simply I<get()>:

 $error = $ENV{error_obj}->get_last_error()
 
Because we are using error stack then at any time we can get full error stack:
 
 @error_stack = $ENV{error_obj}->get_stack()

For detailed description see B<METHODS> section of this manual.


=head2 HOW TO WRITE ANY FUNCTION/METHOD WITH ERROR HANDLING

Each function/method MUST return some result. If no special result 
(like user list, etc.) should be returned - function MUST return B<1> if call 
completed successfully or B<undef> otherwise.

If any error occurs while function execution then function MUST push error 
into error stack.

Error handling in this case will looks like this:

 use HSPC::ErrorStack;
 
 sub test_function
 {
    ...
    if ($local_error) {
        $ENV{error_obj}->set(mod => MOD_CP, smod => APACHE, err => 1);
        return undef;
    } else {
        return 1;
    }
 }
 
 ...
 
 if (test_function()) {
    print "All OK\n";
 } else {
    print "Error occurs: ".$ENV{error_obj}->get_last_error_message();
 }

=head2 ADDING NEW ERROR CODES

=over

=item 1.

Edit HSPC SQL script and add new module and new error codes to database,
for example:

 INSERT INTO module (module_id,name) VALUES (1,"vztools");
 INSERT INTO error (module_id,error_id,text) VALUES(1,1,"Net Error");

Note: If error description and moidule codes not added to database - 
all will work, but error descriptions will be empty.

=back

=head1 IMPLEMENTATION

This chapter describes code structure and database structure.

=head2 ERROR CODE STRUCTURE

See L<hspc-error.pdf> for detail structure of error code and database structure.

There are some methods to get module ID, error ID and error parameters 
from full or extended error code:
I<get_module_id($full_error_code)>, 
I<get_submodule_id($full_error_code)>, 
I<get_error_id($full_error_code)>,
I<get_error_params($full_error_code)>.

Also presented method for make full error code: 
I<make_full_error_code($module_id, $submodule_id, $error_id, $params)>


=head1 METHODS

=cut

#----------------------------------------------------------------------------

use strict;
use HSPC::WebDB;
use Carp;
use HSPC::Localization qw(:DEFAULT string string_lang);
use MIME::Base64;
use HSPC::Error;

use HSPC::Logger qw(sw_die sw_log_debug);
#use vars qw(
#	@ISA @EXPORT
#);
#@ISA = qw (Exporter);
#@EXPORT = qw (load_constants);

#############################################################################

=head2 CONSTRUCTORS

=cut

#############################################################################

=pod

I<new()> - Constructor. No arguments.

=cut

sub new
{
    my $class = shift;
    my $self;
    
    $self->{error_stack} = [];
    
    bless ($self, $class);
    return $self;
}

#----------------------------------------------------------------------------

sub DESTROY
{
    # Close database connection
}

#----------------------------------------------------------------------------

#############################################################################

=head2 GETTING ERROR CODES

=cut

#############################################################################

=pod

I<get_last_error()> - return full last error code from error stack.

=cut

sub get_last_error
{
    my $self = shift;
    
    return $self->{error_stack}->[$#{$self->{error_stack}}];
}

#----------------------------------------------------------------------------

=pod

I<get()> - synonym for I<get_last_error()>.

=cut

sub get
{
    my $self = shift;
    return $self->get_last_error();
}

#----------------------------------------------------------------------------

=pod

I<get_error_message($full_error_code)> - return text description for $full_error_code

=cut

sub get_error_message
{
    my $self = shift;
    my $error_code = shift;

    my $sql;
    my $message;

    return undef unless $error_code;
    my $module_id = $self->get_module_id($error_code);
    my $submodule_id = $self->get_submodule_id($error_code);
    my $error_id  = $self->get_error_id($error_code);
    my $params = $self->get_error_params($error_code);

	## check if we use error_plugin function 
	if ($module_id == &MOD_PLUGIN) {
		sw_die("There is no plugin id!") unless ($params->{__plugin_id__});
		$sql = qq (SELECT text_id FROM plugin_errors WHERE plugin_id = ? 
			AND error_id = ?);
		$message = HSPC::WebDB::select_result(
			$sql, $params->{__plugin_id__}, $error_id
		);
	} else {
		$sql = qq (SELECT text_id FROM error WHERE module_id = ? 
			AND submodule_id = ? AND error_id = ?);
		$message = HSPC::WebDB::select_result(
			$sql, $module_id, $submodule_id, $error_id
		);
	}

    return '' unless ($message);	
    $message = string($message) || $message;
    $message =~ s/\%(\w+)\%/$params->{$1}/g;

    return $message;
}

#----------------------------------------------------------------------------

=pod

I<get_last_error_message()> - return text description for last error code 
added into error stack.

=cut

sub get_last_error_message
{
    my $self = shift;
    my $error_code = $self->get_last_error();
    return undef unless $error_code;

    return $self->get_error_message($error_code);
}

#----------------------------------------------------------------------------

=pod

I<get_stack()> - return full error stack as array of error codes.

=cut

sub get_stack
{
    my $self = shift;
    return @{$self->{error_stack}};
}

#----------------------------------------------------------------------------

## return full error stack as string. String contains
## messages for all errors splitted by separator
## => no_codes - do not add error code
## => separator - separate errors in string with this (default - "\n")

sub get_stack_as_text {
    my $self = shift;
    my %h = (
    	no_codes => undef,
    	separator => "\n",
    	@_
    );
    my $text = '';
    my $code;
    my %codes = ();
    foreach ($self->get_stack) {
        $code = $self->ext2normal_code($_);
        ## skip duplicate entries 
        if($code && !$codes{$code}) {
			$codes{$code} = 1;
			$text .= ($h{no_codes} ? '' :"$code: ")
				.$self->get_error_message($_)
				.$h{separator};
    	}
    }
    
    return $text;
}


#############################################################################

=head2 SETTING ERROR CODES

=cut

#############################################################################

=pod

I<set_full_error ($code)> - push full error code $code into error stack.
To make full error code use I<make_full_error_code()> method.
Otherwise use I<set_last_error()>.

=cut

sub set_full_error
{
    my $self = shift;
    my $full_code = shift;
    
    sw_die(" - full error code undefined!") unless $full_code;
    
    # Log errors
    if ($ENV{system_obj}) {
        sw_log_debug ('[ErrorStack]: '.$self->ext2normal_code($full_code).' '.$self->get_error_message($full_code));
    }
    
    push @{$self->{error_stack}}, $full_code;
}

#----------------------------------------------------------------------------

=pod

I<set_last_error(...)> - push error code into error stack.

=over

=item => mod (integer)
=item => smod (integer)

Module and submodule IDs. Use B<HSPC::ErrorStack> constants as module/submodule ID.

=item => err (integer)

Error code inside module.

=item => params (hashref, optional)

Hash contains parameters names and values for extended error messages.
See B<IMPLEMENTATION> for details.

=back

=cut

sub set_last_error
{
    my $self = shift;
    my %arg = (
        mod  => undef,
        smod => 0,
        err  => undef,
        params    => undef,
        @_
    );
    
    sw_die(" - 'err' parameter undefined") unless defined $arg{err};
    sw_die(" - 'mod' parameter undefined") unless $arg{mod};
    sw_die(" - Invalid parameters type for extended error codes - must be HASH ") if ($arg{params} && ref($arg{params}) ne 'HASH');
    
    return unless $arg{err};
    
    my $full_code = $self->make_full_error_code($arg{mod}, $arg{smod}, $arg{err}, $arg{params});
    
    # Log errors
    if ($ENV{system_obj}) {
        sw_log_debug ('[ErrorStack]: '.$self->ext2normal_code($full_code).' '.$self->get_error_message($full_code));
    }
    
    push @{$self->{error_stack}}, $full_code;
}

#----------------------------------------------------------------------------

=pod

I<set()> - synonym for I<set_last_error()>.

=cut

sub set
{
    my $self = shift;
    
    $self->set_last_error(@_);
}

#############################################################################

=head2 ERROR CODE STRUCTURE RELATED METHODS

=cut

#############################################################################

=pod

I<get_module_id($full_error_code)> - extracts module id from $full_error_code
and return it.

=cut

sub get_module_id
{
    my $self = shift;
    my $full_code = shift;
    $full_code = $self->ext2normal_code($full_code);

    return $full_code >> 16;
}

#----------------------------------------------------------------------------

=pod

I<get_submodule_id($full_error_code)> - extracts submodule id from $full_error_code
and return it.

=cut

sub get_submodule_id
{
    my $self = shift;
    my $full_code = shift;
    $full_code = $self->ext2normal_code($full_code);

    $full_code  = $full_code >> 8;
    return $full_code & 0xFF;
}

#----------------------------------------------------------------------------

=pod

I<get_error_id($full_error_code)> - extracts error id from $full_error_code
and return it.

=cut

sub get_error_id
{
    my $self = shift;
    my $full_code = shift;
    $full_code = $self->ext2normal_code($full_code);

    return $full_code & 0xFF;
}

#----------------------------------------------------------------------------

=pod

I<get_error_params($full_error_code)> - extracts error parameters from $full_error_code
and return hashref with parameter names and values or undef if parameter not present.

=cut

sub get_error_params
{
    my $self = shift;
    my $code = shift;
    my $param_name; my $param_val;
    my %params;
    
    return undef unless $self->is_extended_code($code);
    
    $code =~ s/^(\d+)-(.*)$/$2/;
    my @paars = split ('-', $code);
        
    foreach (@paars) {
        ($param_name, $param_val) = split (':', $_, 2);
        $param_name = decode_base64($param_name);
        $params{$param_name} = decode_base64($param_val);
    }
    
    return \%params;
}

#----------------------------------------------------------------------------

=pod

I<ext2normal_code($extended_code)> - removes parameters (if present) from 
extended error code and return normal error code.

=cut

sub ext2normal_code
{
    my $self = shift;
    my $code = shift;
    
    $code =~ s/^(\d+).*$/$1/;
    
    return $code;
}

#----------------------------------------------------------------------------

=pod
    
I<is_extended_code($code)> - return 1 if $code is extended error code 
and 0 otherwise.

=cut

sub is_extended_code
{
    my $self = shift;
    my $code = shift;
    return 0 unless $code; # Empty code - is not extended code
    
    # Extended code contains non-digits
    return ($code =~ /^\d+$/) ? 0 : 1;
}

#----------------------------------------------------------------------------

=pod

I<make_full_error_code($module_id, $submodule_id, $error_id, $params)> - 
make full error code from given $module_id, $submodule_id, $error_id and $params.

$params is a reference to hash, contains parameter names and values for extended error messages.
Params may be omitted. See B<IMPLEMENTATION> for details.

=cut

sub make_full_error_code
{
    my $self = shift;
    my ($module_id, $submodule_id, $error_id, $params) = @_;
    my $full_code =  $module_id * 0x10000 + $submodule_id * 0x100 + $error_id;
    
    sw_die("Invalid parameters type for extended error codes - must be HASH ") if ($params && ref($params) ne 'HASH');
    
    my $param_name; my $param_val;
    while (($param_name, $param_val) = each %$params) {
        $full_code = $full_code.'-'.encode_base64($param_name, '').':'.encode_base64($param_val, '');
    }

    return $full_code;
}

#############################################################################

=head2 MISC FUNCTIONS

=cut

#############################################################################

=pod

I<clear_last_error()> - deletes last error from error stack and return it's code.

=cut

sub clear_last_error
{
    my $self = shift;
    
    return pop @{$self->{error_stack}};
}

#----------------------------------------------------------------------------

=pod

I<clear_stack()> - deletes all errors from error stack.

=cut

sub clear_stack
{
    my $self = shift;
    $self->{error_stack} = [];
}

#----------------------------------------------------------------------------

=pod

I<is_stack_empty()> - return 1 if error stack is empty and undef otherwise.

=cut

sub is_stack_empty
{
    my $self = shift;
    
    return ($#{$self->{error_stack}} == -1) ? 1 : undef;
}

#----------------------------------------------------------------------------

=head1 NOTES

If you have difficulties on splitting you application into modules or 
organazing errors pool refer to errors keeper: L<olsh@sw.ru>

For technical questions regarding integration of 
B<HSPC::ErrorsStack> into your application refer to module author: L<eaa@sw.ru>

=head1 SEE ALSO

L<HSPC::ErrorFunc(3)>, L<hspc-error.pdf>

=cut

#----------------------------------------------------------------------------

1;
